// Copyright 2019, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Authors: Patrick Brosi <brosi@informatik.uni-freiburg.de>

#include <vector>
#include "osmfixer/server/StatServer.h"
#include "util/Misc.h"
#include "util/String.h"
#include "util/geo/Geo.h"
#include "util/log/Log.h"

using osmfixer::StatServer;
using osmfixer::Params;

// _____________________________________________________________________________
util::http::Answer StatServer::handle(const util::http::Req& req,
                                      int con) const {
  UNUSED(con);
  util::http::Answer a;
  try {
    Params params;
    auto cmd = parseUrl(req.url, req.payload, &params);

    if (cmd == "/") {
      a = util::http::Answer("200 OK", "osmfixer");
    } else if (cmd == "/map") {
      a = handleMapReq(params);
    } else if (cmd == "/stat") {
      a = handleStatReq(params);
    } else {
      a = util::http::Answer("404 Not Found", "dunno");
    }
  } catch (std::runtime_error e) {
    a = util::http::Answer("400 Bad Request", e.what());
  } catch (std::invalid_argument e) {
    a = util::http::Answer("400 Bad Request", e.what());
  } catch (...) {
    a = util::http::Answer("500 Internal Server Error",
                           "Internal Server Error.");
  }

  a.params["Access-Control-Allow-Origin"] = "*";
  a.params["Server"] = "osmfixer";

  return a;
}

// _____________________________________________________________________________
util::http::Answer StatServer::handleMapReq(const Params& pars) const {
  if (pars.count("bbox") == 0 || pars.find("bbox")->second.empty())
    throw std::invalid_argument("No bbox specified.");
  std::string cb;
  if (pars.count("cb")) cb = pars.find("cb")->second.c_str();
  auto box = util::split(pars.find("bbox")->second, ',');

  if (box.size() != 4) throw std::invalid_argument("Invalid request.");

  double lat1 = atof(box[0].c_str());
  double lng1 = atof(box[1].c_str());
  double lat2 = atof(box[2].c_str());
  double lng2 = atof(box[3].c_str());

  std::cout << pars.find("bbox")->second << std::endl;

  util::geo::DBox bbox(util::geo::DPoint(lat1, lng1),
                       util::geo::DPoint(lat2, lng2));

  LOG(INFO) << "Request for bounding box " << util::geo::getWKT(bbox);

  std::stringstream json;

  json << std::setprecision(10);

  if (cb.size()) json << cb << "(";
  json << "{\"stats\":[";

  auto ret = _idx->getStations(bbox);
  char sep = ' ';
  for (auto stat : ret) {
    json << sep;
    sep = ',';
    printStation(stat, &json);
  }

  json << "], \"groups\":[";
  auto gret = _idx->getGroups(bbox);
  sep = ' ';
  for (auto group : gret) {
    if (group->stations.size() < 2) continue;
    json << sep;
    sep = ',';
    printGroup(group, &json);
  }

  json << "]}";

  if (cb.size()) json << ")";

  auto answ = util::http::Answer("200 OK", json.str(), true);
  answ.params["Content-Type"] = "application/javascript; charset=utf-8";

  return answ;
}

// _____________________________________________________________________________
void StatServer::printStation(const Station* stat, std::ostream* out) const {
  auto latLng =
      util::geo::webMercToLatLng<double>(stat->pos.getX(), stat->pos.getY());
  (*out) << "{\"id\":" << stat->id << ",\"lat\":" << latLng.getY()
         << ",\"lon\":" << latLng.getX();

  if (stat->origGroup != stat->group || _idx->getGroup(stat->group)->osmid == 1)
    (*out) << ",\"suggestion\":1";

  if (stat->attrErrs.size())
    (*out) << ",\"attrerrs\":" << stat->attrErrs.size();
  (*out) << "}";
}

// _____________________________________________________________________________
void StatServer::printGroup(const Group* group, std::ostream* out) {
  std::vector<util::geo::DPoint> projPoly;

  for (auto p : group->poly.getOuter()) {
    projPoly.push_back(util::geo::webMercToLatLng<double>(p.getX(), p.getY()));
  }
  (*out) << "{\"id\":" << group->id << ",\"poly\":[";
  char sep = ' ';
  for (auto p : projPoly) {
    (*out) << sep << "[" << p.getY() << "," << p.getX() << "]";
    sep = ',';
  }
  (*out) << "]";
  if (group->osmid == 1) (*out) << ",\"new\":1";
  (*out) << "}";
}

// _____________________________________________________________________________
std::string StatServer::parseUrl(std::string u, std::string pl,
                                 std::map<std::string, std::string>* params) {
  auto parts = util::split(u, '?');

  if (parts.size() > 1) {
    auto kvs = util::split(parts[1], '&');
    for (auto kv : kvs) {
      auto kvp = util::split(kv, '=');
      if (kvp.size() == 1) kvp.push_back("");
      (*params)[util::urlDecode(kvp[0])] = kvp[1];
    }
  }

  // also parse post data
  auto kvs = util::split(pl, '&');
  for (auto kv : kvs) {
    auto kvp = util::split(kv, '=');
    if (kvp.size() == 1) kvp.push_back("");
    (*params)[util::urlDecode(kvp[0])] = kvp[1];
  }

  return util::urlDecode(parts.front());
}

// _____________________________________________________________________________
util::http::Answer StatServer::handleStatReq(const Params& pars) const {
  if (pars.count("id") == 0 || pars.find("id")->second.empty())
    throw std::invalid_argument("No ID specified.");
  std::string cb;
  if (pars.count("cb")) cb = pars.find("cb")->second.c_str();

  size_t sid = atol(pars.find("id")->second.c_str());
  auto stat = _idx->getStation(sid);

  if (!stat) return util::http::Answer("404 Not Found", "Station not found.");

  std::stringstream json;

  json << std::setprecision(10);

  if (cb.size()) json << cb << "(";
  json << "{\"id\":" << sid << ","
       << "\"osmid\":" << stat->osmid << ","
       << "\"attrs\":{";

  char sep = ' ';

  for (const auto& par : stat->attrs) {
    json << sep;
    sep = ',';
    json << "\"" << util::jsonStringEscape(par.first) << "\":[";

    char sep2 = ' ';
    for (const auto& val : par.second) {
      json << sep2;
      sep2 = ',';
      json << "\"" << util::jsonStringEscape(val) << "\"";
    }
    json << "]";
  }

  json << "},\"attrerrs\":[";

  sep = ' ';
  for (const auto& err : stat->attrErrs) {
    json << sep;
    sep = ',';
    const auto otherStat = _idx->getStation(err.otherId);
    json << "{";
    json << "\"attr\":[\"" << err.attr << "\",\"" << stat->attrs.at(err.attr)[0]
         << "\"]";
    json << ",\"other_attr\":[\"" << err.otherAttr << "\",\""
         << otherStat->attrs.at(err.otherAttr)[0] << "\"]";
    json << ",\"conf\":" << err.conf;
    json << ",\"other\":" << err.otherId;
    json << ",\"other_osmid\":" << otherStat->osmid;
    json << "}";
  }

  json << "]";

  json << ",\"suggestions\":[";

  // suggestions
  if (stat->group != stat->origGroup || _idx->getGroup(stat->group)->osmid == 1) {
    if (_idx->getGroup(stat->origGroup)->osmid < 2) {
      if (_idx->getGroup(stat->group)->osmid == 1) {
        json << "{\"type\": 1, \"target_gid\":" << stat->group << "}";
      } else if (_idx->getGroup(stat->group)->osmid > 1) {
        json << "{\"type\": 2, \"target_gid\":" << stat->group
             << ",\"target_osm_rel_id\":" << _idx->getGroup(stat->group)->osmid
             << "}";
      }
    } else {
      if (_idx->getGroup(stat->group)->osmid == 1) {
        json << "{\"type\": 3,\"orig_gid\":" << stat->origGroup
             << ",\"orig_osm_rel_id\":"
             << _idx->getGroup(stat->origGroup)->osmid
             << ",\"target_gid\":" << stat->group << "}";
      } else if (_idx->getGroup(stat->group)->osmid > 1) {
        json << "{\"type\": 4,\"orig_gid\":" << stat->origGroup
             << ",\"orig_osm_rel_id\":"
             << _idx->getGroup(stat->origGroup)->osmid
             << ",\"target_gid\":" << stat->group
             << ",\"target_osm_rel_id\":" << _idx->getGroup(stat->group)->osmid
             << "}";
      } else {
        json << "{\"type\": 5,\"orig_gid\":" << stat->origGroup
             << ",\"orig_osm_rel_id\":"
             << _idx->getGroup(stat->origGroup)->osmid
             << ",\"target_gid\":" << stat->group
             << ",\"target_osm_rel_id\":" << _idx->getGroup(stat->group)->osmid
             << "}";
      }
    }
  }

  json << "]";

  json << "}";

  if (cb.size()) json << ")";

  auto answ = util::http::Answer("200 OK", json.str(), true);
  answ.params["Content-Type"] = "application/javascript; charset=utf-8";

  return answ;
}
